@@ -2,6 +2,8 @@
 #define VQSORT_ONLY_STATIC 1
 #include "hwy/contrib/sort/vqsort-inl.h"
 
+#if VQSORT_ENABLED
+
 #define DISPATCH_VQSORT(TYPE) \
 template<> void NPY_CPU_DISPATCH_CURFX(QSort)(TYPE *arr, intptr_t size) \
 { \
@@ -18,3 +20,5 @@ namespace np { namespace highway { namespace qsort_simd {
     DISPATCH_VQSORT(float)
 
 } } } // np::highway::qsort_simd
+
+#endif // VQSORT_ENABLED
